"""
Defines the module's version, read from the version.txt file.
"""

import os

with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), 'version.txt'),
    'r'
) as f:
    __version__ = f.read().strip()



"""
Defines the (de-)serialization for special built-in types.
"""

from types import SimpleNamespace
from numbers import Complex
from functools import singledispatch
from collections.abc import Iterable, Mapping, Hashable

from ._base_classes import Deserializable

from ._save_load import from_hdf5, to_hdf5, to_hdf5_singledispatch
from ._subscribe import subscribe_hdf5, TYPE_TAG_KEY

__all__ = []


class _SpecialTypeTags(SimpleNamespace):
    """
    Defines the type tags for special types.
    """
    LIST = 'builtins.list'
    TUPLE = 'builtins.tuple'
    DICT = 'builtins.dict'
    NUMBER = 'builtins.number'
    STR = 'builtins.str'
    NONE = 'builtins.none'


@subscribe_hdf5(_SpecialTypeTags.DICT)
class _DictDeserializer(Deserializable):
    """Helper class to de-serialize dict objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        try:
            items = from_hdf5(hdf5_handle['items'])
            return {_ensure_hashable(k): v for k, v in items}
        # Handle legacy dicts with only string keys:
        except KeyError:
            res = dict()
            value_group = hdf5_handle['value']
            for key in value_group:
                res[key] = from_hdf5(value_group[key])
            return res


@subscribe_hdf5(_SpecialTypeTags.LIST)
class _ListDeserializer(Deserializable):
    """Helper class to de-serialize list objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return _deserialize_iterable(hdf5_handle)


@subscribe_hdf5(_SpecialTypeTags.TUPLE)
class _TupleDeserializer(Deserializable):
    """Helper class to de-serialize tuple objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return tuple(_deserialize_iterable(hdf5_handle))


def _deserialize_iterable(hdf5_handle):
    int_keys = [key for key in hdf5_handle if key != TYPE_TAG_KEY]
    return [from_hdf5(hdf5_handle[key]) for key in sorted(int_keys, key=int)]


@subscribe_hdf5(_SpecialTypeTags.NUMBER, extra_tags=(_SpecialTypeTags.STR, ))
class _ValueDeserializer(Deserializable):
    """Helper class to de-serialize numbers and strings objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return hdf5_handle['value'].value


@subscribe_hdf5(_SpecialTypeTags.NONE)
class _NoneDeserializer(Deserializable):
    """Helper class to de-serialize ``None``."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return None


def add_type_tag(tag):
    """
    Decorator which adds the given type tag when creating the HDF5 object.
    """

    def outer(func):
        def inner(obj, hdf5_handle):
            hdf5_handle[TYPE_TAG_KEY] = tag
            func(obj, hdf5_handle)

        return inner

    return outer


@to_hdf5_singledispatch.register(Iterable)
@add_type_tag(_SpecialTypeTags.LIST)
def _(obj, hdf5_handle):
    _serialize_iterable(obj, hdf5_handle)


@to_hdf5_singledispatch.register(tuple)
@add_type_tag(_SpecialTypeTags.TUPLE)
def _(obj, hdf5_handle):
    _serialize_iterable(obj, hdf5_handle)


def _serialize_iterable(obj, hdf5_handle):
    for i, part in enumerate(obj):
        sub_group = hdf5_handle.create_group(str(i))
        to_hdf5(part, sub_group)


@to_hdf5_singledispatch.register(Mapping)
@add_type_tag(_SpecialTypeTags.DICT)
def _(obj, hdf5_handle):
    items_group = hdf5_handle.create_group('items')
    to_hdf5(obj.items(), items_group)


@to_hdf5_singledispatch.register(Complex)
@add_type_tag(_SpecialTypeTags.NUMBER)
def _(obj, hdf5_handle):
    _value_serializer(obj, hdf5_handle)


@to_hdf5_singledispatch.register(str)
@add_type_tag(_SpecialTypeTags.STR)
def _(obj, hdf5_handle):
    _value_serializer(obj, hdf5_handle)


@to_hdf5_singledispatch.register(type(None))
@add_type_tag(_SpecialTypeTags.NONE)
def _(obj, hdf5_handle):
    pass


def _value_serializer(obj, hdf5_handle):
    hdf5_handle['value'] = obj


def _ensure_hashable(obj):
    if isinstance(obj, Hashable):
        return obj
    return _to_hashable(obj)


@singledispatch
def _to_hashable(obj):
    raise ValueError(
        "Cannot convert object '{}' of type '{}' to a hashable object.".format(
            obj, type(obj)
        )
    )


@_to_hashable.register(Iterable)
def _(obj):
    return tuple(obj)



"""
Defines the mapping between type tags and serializable classes.
"""

from decorator import decorator

from fsc.export import export

SERIALIZE_MAPPING = {}
TYPE_TAG_KEY = 'type_tag'


@export
def subscribe_hdf5(type_tag, extra_tags=(), check_on_load=True):
    """
    Class decorator that subscribes the class with the given type_tag for serialization.

    :param type_tag: Unique identifier of the class, which is injected into the HDF5 data to identify the class.
    :type type_tag: str

    :param extra_tags: Additional tags which should be deserialized to the given class.
    :type extra_tags: tuple(str)

    :param check_on_load: Flag that determines whether the 'type_tag' is checked when de-serializing the object.
    :type check_on_load: bool
    """

    def inner(cls):
        all_type_tags = [type_tag] + list(extra_tags)
        for tag in all_type_tags:
            if tag in SERIALIZE_MAPPING:
                raise ValueError(
                    "The given type_tag '{}' exists already in the SERIALIZE_MAPPING"
                    .format(tag)
                )
            SERIALIZE_MAPPING[tag] = cls

        if hasattr(cls, 'to_hdf5'):

            @decorator
            def set_type_tag(to_hdf5_func, self, hdf5_handle, *args, **kwargs):
                if TYPE_TAG_KEY not in hdf5_handle:
                    hdf5_handle[TYPE_TAG_KEY] = type_tag
                else:
                    assert isinstance(self, cls)
                return to_hdf5_func(self, hdf5_handle, *args, **kwargs)

            cls.to_hdf5 = set_type_tag(cls.to_hdf5)  # pylint: disable=no-value-for-parameter

        if check_on_load:

            @decorator
            def check_type_tag(
                from_hdf5_func, curr_cls, hdf5_handle, *args, **kwargs
            ):
                # check only the top-level class.
                if curr_cls == cls:
                    assert hdf5_handle[TYPE_TAG_KEY].value in all_type_tags
                return from_hdf5_func(curr_cls, hdf5_handle, *args, **kwargs)

            cls.from_hdf5 = classmethod(check_type_tag(cls.from_hdf5.__func__))  # pylint: disable=no-value-for-parameter
        else:
            cls.from_hdf5 = classmethod(cls.from_hdf5.__func__)
        return cls

    return inner



"""
Defines free functions to serialize / deserialize bands-inspect objects to HDF5.
"""

from functools import singledispatch

import h5py
from fsc.export import export

from ._subscribe import SERIALIZE_MAPPING, TYPE_TAG_KEY

__all__ = ['save', 'load']


@export
def from_hdf5(hdf5_handle):
    """
    Deserializes the given HDF5 handle into an object.

    :param hdf5_handle: HDF5 location where the serialized object is stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    try:
        type_tag = hdf5_handle[TYPE_TAG_KEY].value
    except KeyError as err:
        raise ValueError(
            "HDF5 object '{}' cannot be de-serialized: No type information given."
            .format(hdf5_handle.name)
        ) from err
    try:
        obj_class = SERIALIZE_MAPPING[type_tag]
    except KeyError as err:
        raise KeyError(
            "Unknown {} '{}'. The module defining this class needs to be imported before de-serializing the object."
            .format(TYPE_TAG_KEY, type_tag)
        ) from err
    return obj_class.from_hdf5(hdf5_handle)


@export
def to_hdf5(obj, hdf5_handle):
    """
    Serializes a given object to HDF5 format.

    :param obj: Object to serialize.

    :param hdf5_handle: HDF5 location where the serialized object gets stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    if hasattr(obj, 'to_hdf5'):
        obj.to_hdf5(hdf5_handle)
    else:
        to_hdf5_singledispatch(obj, hdf5_handle)


@export
@singledispatch
def to_hdf5_singledispatch(obj, hdf5_handle):  # pylint: disable=unused-argument
    """
    Singledispatch function which is called to serialize and object when it does not have a ``to_hdf5`` method.

    :param obj: Object to serialize.

    :param hdf5_handle: HDF5 location where the serialized object gets stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    raise TypeError(
        "Cannot serialize object '{}' of type '{}'".format(obj, type(obj))
    )


@export
def from_hdf5_file(hdf5_file):
    """
    Loads the object from a file in HDF5 format.

    :param hdf5_file: Path of the file.
    :type hdf5_file: str
    """
    with h5py.File(hdf5_file, 'r') as f:
        return from_hdf5(f)


load = from_hdf5_file  # pylint: disable=invalid-name
load.__doc__ = """Alias for :func:`from_hdf5_file`."""


@export
def to_hdf5_file(obj, hdf5_file):
    """
    Saves the object to a file, in HDF5 format.

    :param obj: The object to be saved.

    :param hdf5_file: Path of the file.
    :type hdf5_file: str
    """
    with h5py.File(hdf5_file, 'w') as f:
        to_hdf5(obj, f)


save = to_hdf5_file  # pylint: disable=invalid-name
save.__doc__ = """Alias for :func:`to_hdf5_file`."""



"""
This module contains functions to save and load objects, using the HDF5 format.
"""

from ._version import __version__

from ._save_load import *
from ._base_classes import *
from ._subscribe import *
# Needs to be loaded on import to define the special type serialization.
from ._simple_mapping import *
from ._special_types import *

__all__ = _save_load.__all__ + _base_classes.__all__ + _subscribe.__all__ + _simple_mapping.__all__  # pylint: disable=undefined-variable



"""
Implements a base class for serializing a given list of attributes of an object.
"""

from fsc.export import export

from ._base_classes import HDF5Enabled
from ._save_load import to_hdf5 as _global_to_hdf5, from_hdf5 as _global_from_hdf5


@export
class SimpleHDF5Mapping(HDF5Enabled):
    """
    Base class for data classes which simply map their member to HDF5 values / groups.

    The child class needs to define a list ``HDF5_ATTRIBUTES`` of attributes which
    should be serialized. The name of the attributes must correspond to the
    name accepted by the constructor.
    """
    HDF5_ATTRIBUTES = ()
    HDF5_OPTIONAL = ()

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        kwargs = dict()
        for key in cls.HDF5_ATTRIBUTES:
            try:
                hdf5_obj = hdf5_handle[key]
            except KeyError:
                if key in cls.HDF5_OPTIONAL:
                    continue
                raise
            try:
                kwargs[key] = hdf5_obj.value
            except AttributeError:
                kwargs[key] = _global_from_hdf5(hdf5_obj)
        return cls(**kwargs)

    def to_hdf5(self, hdf5_handle):
        for key in self.HDF5_ATTRIBUTES:
            try:
                value = getattr(self, key)
            except AttributeError:
                if key in self.HDF5_OPTIONAL:
                    continue
                raise
            try:
                hdf5_handle[key] = value
            except TypeError:
                _global_to_hdf5(value, hdf5_handle.create_group(key))



"""
Usage: pip install .[dev]
"""

import sys

from setuptools import setup

PKGNAME = 'hdf5_io'
PKGNAME_QUALIFIED = 'fsc.' + PKGNAME

with open('doc/description.txt', 'r') as f:
    DESCRIPTION = f.read()
try:
    with open('doc/README', 'r') as f:
        README = f.read()
except IOError:
    README = DESCRIPTION

with open('version.txt', 'r') as f:
    VERSION = f.read().strip()

if sys.version_info < (3, ):
    raise ValueError('only Python 3.x and higher are supported')

setup(
    name=PKGNAME_QUALIFIED,
    version=VERSION,
    packages=[PKGNAME_QUALIFIED],
    url='http://frescolinogroup.github.io/frescolino/pyhdf5io/' +
    '.'.join(VERSION.split('.')[:2]),
    include_package_data=True,
    author='C. Frescolino',
    author_email='frescolino@lists.phys.ethz.ch',
    description=DESCRIPTION,
    install_requires=['numpy', 'decorator', 'h5py', 'fsc.export'],
    extras_require={
        'dev':
        ['pytest', 'pytest-cov', 'yapf==0.24.0', 'prospector==1.1.3', 'pre-commit', 'pylint==2.1.1', 'sphinx', 'sphinx-rtd-theme', 'ipython>=6.2', 'matplotlib']
    },
    long_description=README,
    classifiers=[ # yapf: disable
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities'
    ],
    license='Apache',
)



"""
Base classes for serializing bands-inspect data types.
"""

import abc

import h5py

from fsc.export import export


class Deserializable(abc.ABC):
    """
    Base class for data which can be deserialized from the HDF5 format.
    """

    @classmethod
    @abc.abstractmethod
    def from_hdf5(cls, hdf5_handle):
        """
        Deserializes the object stored in HDF5 format.
        """
        raise NotImplementedError

    @classmethod
    def from_hdf5_file(cls, hdf5_file, *args, **kwargs):
        """
        Loads the object from a file in HDF5 format.

        :param hdf5_file: Path of the file.
        :type hdf5_file: str
        """
        with h5py.File(hdf5_file, 'r') as f:
            return cls.from_hdf5(f, *args, **kwargs)


class Serializable(abc.ABC):
    """
    Base class for data which can be serialized to the HDF5 format.
    """

    @abc.abstractmethod
    def to_hdf5(self, hdf5_handle):
        """
        Serializes the object to HDF5 format, attaching it to the given HDF5 handle (might be a HDF5 File or Dataset).
        """
        raise NotImplementedError

    def to_hdf5_file(self, hdf5_file):
        """
        Saves the object to a file, in HDF5 format.

        :param hdf5_file: Path of the file.
        :type hdf5_file: str
        """
        from ._save_load import to_hdf5_file
        to_hdf5_file(self, hdf5_file)


@export  # pylint: disable=abstract-method
class HDF5Enabled(Serializable, Deserializable):
    """
    Base class for data which can be serialized to and deserialized from HDF5.
    """
    pass
